"""
Handles creation of genomes, either from scratch or by sexual or
asexual reproduction from parents.
"""
from __future__ import division

import math
import random
from itertools import count

from .config import ConfigParameter, DefaultClassConfig
from .math_util import mean
from .six_util import iteritems, itervalues

# TODO: Provide some sort of optional cross-species performance criteria, which
# are then used to control stagnation and possibly the mutation rate
# configuration. This scheme should be adaptive so that species do not evolve
# to become "cautious" and only make very slow progress.


class DefaultReproduction(DefaultClassConfig):
    """
    Implements the default NEAT-python reproduction scheme:
    explicit fitness sharing with fixed-time species stagnation.
    """

    @classmethod
    def parse_config(cls, param_dict):
        return DefaultClassConfig(param_dict,
                                  [ConfigParameter('elitism', int, 0),
                                   ConfigParameter('survival_threshold', float, 0.2),
                                   ConfigParameter('min_species_size', int, 2)])

    def __init__(self, config, reporters, stagnation):
        # pylint: disable=super-init-not-called
        self.reproduction_config = config
        self.reporters = reporters
        self.genome_indexer = count(1)
        self.stagnation = stagnation
        #self.ancestors = {}

    def create_new(self, genome_type, genome_config, num_genomes):
        new_genomes = {}
        for i in range(num_genomes):
            key = next(self.genome_indexer)
            g = genome_type(key)
            g.configure_new(genome_config)
            new_genomes[key] = g
            #self.ancestors[key] = tuple()

        return new_genomes

    @staticmethod
    def compute_spawn(adjusted_fitness, previous_sizes, pop_size, min_species_size):
        """Compute the proper number of offspring per species (proportional to fitness).
        """
        # suma ajd_fitness gatunkow
        af_sum = sum(adjusted_fitness)

        spawn_amounts = []
        # dla adj_fitness i poprzedniej wielkosci kazdego gatunku
        for af, ps in zip(adjusted_fitness, previous_sizes):
            """
            TAK BYLO W BIBLIOTECE
            ZMIENIAM BO: af_sum nigdy nie bedzie < 0 bo af są skalowane do [0,1]
            CHYBA ŻE
            AF_SUM == 0, CZY TO SIE MOŻE ZDARZYC?
            odp. może się zdarzyć gdy funcje przystosowania wszystkich osobników będą
            równe 0
            w moim zadaniu praktycznie nie mozliwe, ale zostawię ten if

            #jeżeli chce sie zmieniać to należy cały if zastąpic:
            s = max(min_species_size, af / af_sum * pop_size)
            """
            if af_sum > 0:  # jezeli suma adj_fit jest dodatnia
                # to zmienna pomocnicza s to max z minimalnej ilosci osobnikow
                # w gatunku i liniowego skalowania ajd_fit do pop_size:
                # czyli liczba dzieci jest wprost propocjonalna do przystosowania gatunku
                # i wielkosci populacji
                s = max(min_species_size, af / af_sum * pop_size)
            else:
                # tu wejdziemy tylko gdy wszystki gatunki maja przystosowanie == 0
                s = min_species_size

            # TODO: tutaj przeprowadz testy dla obu
            ###################################################################################
            #           MOJA WERSJA:
            # DZIAŁA SZYBCIEJ,
            # NIE HAMUJE ZMIAN WIELKOSCI GATUNKÓW
            # JEZELI CHCESZ ZACHOWAĆ PEWNĄ INERCJE WIELKOSCI GATUNKOW, UZYJ TEGO CO BYLO
            # (niżej jest w komentarzu)
            ###################################################################################
            spawn_amounts.append(int(round(s)))
            ###################################################################################
            #       WERSJA KTORA BYŁA W BIBLIOTECE:
            ###################################################################################
            """
            # magic d = (liczba potomkow - poprzedni rozmiar)/2
            # czyli połowa z roznicy wielkosci gatunku przed chwilą i za chwilę
            # TO 0.5 hamuje zmiany wielkosci gatunków
            d = (s - ps) * 0.5
            # zaokrąglij do int
            c = int(round(d))
            spawn = ps  # liczba potomków taka sama jak wielkosc gatunku
            # if abs(c) > 0:  # jezeli roznica wielkosci nie jest zerowa
            # abs(c) > 0 zawraca TRUE zawsze gdy c != 0, a FALSE jedynie gdy c == 0, wiec:
            if c:
                spawn += c  # powieksz wielkosc gatunku o liczbe c
            elif d > 0:
                spawn += 1  # jezeli d jest w [0,1] to zwieksz wielkosc gatunku o 1
            elif d < 0:
                spawn -= 1  # jezeli d jest w [-1,0] do pomniejsz gatunek o 1
            # jak c == 0  to nie zmieniaj wielkosci gatunku

            spawn_amounts.append(spawn)
            """
        # Normalize the spawn amounts so that the next generation is roughly
        # the population size requested by the user.
        total_spawn = sum(spawn_amounts)
        norm = pop_size / total_spawn
        spawn_amounts = [max(min_species_size, int(round(n * norm)))
                         for n in spawn_amounts]

        return spawn_amounts

    def reproduce(self, config, species, pop_size, generation):
        """
        Handles creation of genomes, either from scratch or by sexual or
        asexual reproduction from parents.
        """

        # Filter out stagnated species, collect the set of non-stagnated
        # species members, and compute their average adjusted fitness.
        # The average adjusted fitness scheme (normalized to the interval
        # [0, 1]) allows the use of negative fitness values without
        # interfering with the shared fitness scheme.
        all_fitnesses = []  # fintessy wszystkich osobników nie podlegających stagnacji
        remaining_species = []  # lista przetrwających gatuków
        #   id gatunku, gatunek, flaga stagnacji
        for stag_sid, stag_s, stagnant in self.stagnation.update(species, generation):
            if stagnant:  # jezeli gatunek jest w stagnacji
                self.reporters.species_stagnant(stag_sid, stag_s)
            else:  # jezeli gatunek NIE jest w stagnacji
                # dodaj fintesy osobnikow gatunku do listy pomocnieczej all_fitnesses
                all_fitnesses.extend(m.fitness for m in itervalues(stag_s.members))
                # dodaj gatunek do listy gatunkow nie podlegających stagnacji,
                # czyli przetrawających
                remaining_species.append(stag_s)
        # The above comment was not quite what was happening - now getting fitnesses
        # only from members of non-stagnated species.

        # No species left.
        if not remaining_species:  # jeżeli żaden gatunek nie przetrwał
            species.species = {}
            return {}  # was [] #zwrób pusty słownik

        # Find minimum/maximum fitness across the entire population, for use in
        # species adjusted fitness computation.
        min_fitness = min(all_fitnesses)
        max_fitness = max(all_fitnesses)
        # Do not allow the fitness range to be zero, as we divide by it below.
        # TODO: The ``1.0`` below is rather arbitrary, and should be configurable.
        fitness_range = max_fitness - min_fitness
        if fitness_range == 0:
            fitness_range = 1
        for afs in remaining_species:  # dla kazdego przeżywajacego gatunku
            # Compute adjusted fitness.
            # fitness gatunku przezywajacego od którego zależec będzie liczba
            # jego potomków; to jest od razu suma z fi z artykulu Stanley'a
            # msf = suma(fi) / liczbe osobnikow w gatunku:
            # Every species is assigned a potentially different number of offspring in
            # proportion to the sum of ad- justed fitnesses fi of its member organisms.
            msf = mean([m.fitness for m in itervalues(afs.members)])
            # adjusted_fitness gatunku jest propoorcjonalny do msf
            # jezeli gatunek bedzie mial blisko fitness najlepszego osobnika
            # w populacji to afs.adjuste_fitness bedzie blisko 1
            # jezeli gatunek bedzie blisko najsłabszego osobnika w populacji
            # to afs.adjusted_fitness bedzie blisko 0
            # czyli to jest skalowanie - NORMALIZOWANIE przystosowania gatunku do [0,1]
            afs.adjusted_fitness = (msf - min_fitness) / fitness_range

        # lista adj_fitness gatunkow
        adjusted_fitnesses = [s.adjusted_fitness for s in remaining_species]
        # srednia z adj_fitness gatunków
        #avg_adjusted_fitness = mean(adjusted_fitnesses)
        #self.reporters.info(
        #   "Average adjusted fitness: {:.3f}".format(avg_adjusted_fitness))

        # Compute the number of new members for each species in the new generation.
        previous_sizes = [len(s.members)
                          for s in remaining_species]  # liczebnosci gatunkow
        # minimalna ilosc osobnikow w gatunku
        min_species_size = max(self.reproduction_config.min_species_size,
                               self.reproduction_config.elitism)
        # oblicz liczbe potomkow dla kazdego gatunku
        spawn_amounts = self.compute_spawn(adjusted_fitnesses, previous_sizes,
                                           pop_size, min_species_size)
        new_population = {}
        species.species = {}  # wyczysć słownik key gatunków
        newNodesInGen = {}  # słownik mutacji zachodzących w danej generacji
        # dla liczby potomków (spawn) każdego przetrawjącego gatunku (s)
        for spawn, s in zip(spawn_amounts, remaining_species):
            # If elitism is enabled, each species always at least gets to retain its elites.
            spawn = max(spawn, self.reproduction_config.elitism)
            assert spawn > 0

            # The species has at least one member for the next generation, so retain it.
            # przepisz stare osobniki gatunku
            old_members = list(iteritems(s.members))
            # aktualne osobniki gatunku, na razie pusto
            s.members = {}
            # dodaj gatunek do listy gatunku
            # s.key pełni role indeksu, jak key w MySQL
            species.species[s.key] = s

            # Sort members in order of descending fitness.
            old_members.sort(reverse=True, key=lambda x: x[1].fitness)

            # Transfer elites to new generation.
            if self.reproduction_config.elitism > 0:
                for i, m in old_members[:self.reproduction_config.elitism]:
                    new_population[i] = m
                    spawn -= 1

            if spawn <= 0:
                continue

            # Only use the survival threshold fraction to use as parents for the next generation.
            # Do romnażania weź jedynie survival_threshold (np. 0.2 = 20%) osobników z gatunku
            repro_cutoff = int(math.ceil(self.reproduction_config.survival_threshold *
                                         len(old_members)))
            # ale nie mniej niż dwoje rodziców
            # Use at least two parents no matter what the threshold fraction result is.
            repro_cutoff = max(repro_cutoff, 2)
            old_members = old_members[:repro_cutoff]

            # Randomly choose parents and produce the number of offspring allotted to the species.
            # dopóki liczba potomków nie wyczerpała się
            while spawn > 0:
                spawn -= 1
                # losowo wylosuj rodziców z zbioru rodziców
                # możliwe jest ponowne losowanie tego samego osobnika
                # osobnik może nie być w ogóle wylosowany
                parent1_id, parent1 = random.choice(old_members)
                parent2_id, parent2 = random.choice(old_members)

                # Note that if the parents are not distinct, crossover will produce a
                # genetically identical clone of the parent (but with a different ID).
                # stwórz dziecko z kolejnym numerem
                gid = next(self.genome_indexer)
                # tworzy dzieciaka, ale go nie konfiguruje
                child = config.genome_type(gid)
                # krzyżuj dziecko z rodziców
                child.configure_crossover(parent1, parent2, config.genome_config)
                # mutuj dzieciaka
                child.mutate(config.genome_config, newNodesInGen)
                # dodaj dziecko do nowej populacji
                new_population[gid] = child
                # pamięta rodziców każdego osobnika (nawet usuwanych)
                # rosnie w nieskonczonosc
                # gdzie to jest używane do homologi?
                #self.ancestors[gid] = (parent1_id, parent2_id)

        return new_population
