import numpy as np
import matplotlib.pyplot as plt


# pokaz srednia z srednich populacji z wszystkich run
def plot_average_of_average(stats, hmRun, hmGen, figID):
    avgFitRun = np.zeros((hmRun, hmGen))

    for i, s in enumerate(stats):
        run_gens_mean = np.array(s.get_fitness_mean())
        avgFitRun[i, :] = run_gens_mean
    avgFitRun = np.mean(avgFitRun, axis=0)
    fig = plt.figure(figID)
    ax = fig.gca()
    plt.plot(range(0, hmGen, 1), avgFitRun)
    ax.xaxis.get_major_locator().set_params(integer=True)
    plt.xlabel(r"Generations $[\#]$")
    plt.ylabel(r"Average population fitness from all algorithm calls $[-]$")
    plt.show()
    fig.savefig('srednie_z_gens.png')


# pokaz sredni fit najlepszych ze wszystkich run
def plot_average_of_bests(stats, hmRun, hmGen, figID):
    bestsFitRun = np.zeros((hmRun, hmGen))
    bestsMSERun = np.zeros((hmRun, hmGen))
    for i, s in enumerate(stats):
        run_bests_fit = np.array([c.fitness for c in s.most_fit_genomes])
        run_bests_MSE = np.array([c.MSE for c in s.most_fit_genomes])
        bestsFitRun[i, :] = run_bests_fit
        bestsMSERun[i, :] = run_bests_MSE
    avgBestsFit = np.mean(bestsFitRun, axis=0)
    avgBestsMSE = np.mean(bestsMSERun, axis=0)

    fig = plt.figure(figID)

    ax1 = fig.gca()
    ax1.xaxis.get_major_locator().set_params(integer=True)
    ax1.set_ylabel(r"Average best individual fitness from all algorithm calls $[-]$")
    plt.xlabel(r"Generations $[\#]$")
    plt.plot(avgBestsFit)

    color = 'tab:red'
    ax2 = ax1.twinx()
    ax2.set_ylabel(r'Average best individual MSE from all algorithm calls $[-]$', color=color)
    plt.plot(avgBestsMSE, color=color)
    ax2.tick_params(axis='y', labelcolor=color)
    fig.tight_layout()

    plt.show()
    fig.savefig('srednie_z_bests.png')
