from __future__ import print_function
import os
import scipy.io
import neatSource as neat
import visualize
import numpy as np
import matplotlib.pyplot as plt
import time
import plot_z_hmRun as ploty
import itertools
import pickleData as pkl


plt.rcParams['text.usetex'] = True
plt.rcParams['font.size'] = 22

PWR = True
useParallel = True
hmGen = 2500
hmWorkers = 12
hmRun = 10

# dodać, że wagi są zmieniane najbardziej dla najgorszych
# zobaczyc
if PWR:
    # dane
    mat = scipy.io.loadmat('daneFinal.mat')
    Pav = mat["Pav"]
    u = mat["x"]
    # skalowanie danych
    nominalP = 3436*10**6
    hmSample = 1100
    Pav = Pav / nominalP
    target = Pav

    y0 = target[0]  # nominalna moc
    u0 = -1.098  # nominalne zagłębienie
    t = np.arange(0, hmSample, 1)
else:
    def f(u, y):
        return -0.05*y[1] + 0.02*y[5] + np.sin(y[10]/10) + u[15]
    hmSample = 1100
    t = np.linspace(0, hmSample, hmSample)
    # u = np.sin(t/30)
    # u[200:400] = 1
    # u[400: 600] = 0
    # u[600:800] = -1
    # u[800:1100] = 1
    u = np.zeros(t.shape)
    u[0:100] = np.linspace(0, 1, 100)
    u[100:125] = 1
    u[125:200] = np.linspace(1, -0.10, 75)
    u[200:400] = -0.1 - 0.5*np.sin(np.linspace(0, 3.14, 200))
    u[400:500] = -0.5
    u[500:600] = 0.25*np.sin(np.linspace(0, 3.14*2, 100))
    u[600:650] = np.linspace(0, 0.5, 50) + 0.25*np.sin(np.linspace(0, 2*3.14, 50))
    u[650:700] = 0.5 + 0.25*np.sin(np.linspace(0, 2*3.14, 50))
    u[700:1000] = np.linspace(0.5, 0, 300) + np.linspace(0.25, 0, 300)*np.sin(np.linspace(0, 6*3.14, 300))
    u[1000:1100] = 0
    y = np.zeros(t.shape)
    maxD = 15+1
    for k in range(maxD, hmSample):
        y[k] = y[k-1] + f(u[k:k-maxD:-1], y[k:k-maxD:-1])

    u = np.array(u[maxD-1:]).reshape((-1, 1))
    target = np.array(y[maxD-1:]).reshape((-1, 1))
    target = target/30.0  # (target - min(target))*2 / (max(target)-min(target)) - 1
    t = np.array(t[maxD-1:])
    hmSample -= maxD
    y0 = y[0]
    u0 = 0

plt.plot(u)
plt.plot(target)
plt.show()

def netOutput(net, u):  # na razie tylko dla SISO bo tak szybciej
    output = [0.0]*len(u)
    for idx, u_k in enumerate(u):  # w kolejnych chwilach, chwil tyle co wymuszenia
        output[idx] = net.activate(u_k)[0]
    return output


def eval_genome(genome, config):
    net = neat.FeedForwardNetwork.create(genome, config)
    # uwzględnij du i dy
    output = np.array(netOutput(net, u)).reshape((-1, 1))
    MSE = np.sum((output - target)**2) / hmSample
    # NinNet = len(net.node_evals) - 1  # -1 bo nie biorę wyjsciowego
    fit = -1000*MSE  # - 0.01*NinNet
    if np.isnan(MSE) or np.isnan(fit) or MSE > 10000:
        fit = -10000
        MSE = 10000
    return fit, MSE


def eval_parallel(genome, config):
    return eval_genome(genome, config)


def eval_notParallel(genomes, config):
    for genome_id, genome in genomes:
        genome.fitness, genome.MSE = eval_genome(genome, config)


def run(config_file):
    # startTime = time.time()
    # Load configuration.
    # w konfiguracji podajesz wskaźniki na classy opisujące:
    # gen, reprodukcje, gatunek, stagnacje i plik w ktorym są ich parametry
    # konfiguracja zapisuje klasy do obiektu config oraz pobiera ich
    # konfiguracje z pliku config_file oraz zapisuje to do atrybutów
    # konfiguracji
    config = neat.Config(neat.DefaultGenome, neat.DefaultReproduction,
                          neat.DefaultSpeciesSet, neat.DefaultStagnation,
                          config_file)

    # Create the population, which is the top-level object for a NEAT run.
    # jest możliwosc wgrania populacji początkowej
    p = neat.Population(config)

    # Add a stdout reporter to show progress in the terminal.
    # to jest reporter, który printuje info o danej generacji w konsoli
    #p.add_reporter(neat.StdOutReporter(True))
    # to jest statystyczny reporter który zbiera (w RAM) różne ciekawe info
    # podczas działania algorytmu
    # TODO: NIE ZAPISUJE AUTOMATYCZNIE DO PLIKU, ALE MOŻNA TO ŁATWO DODAĆ
    # WARTO DODAĆ ZAPIS DO PLIKU, ABY ZWOLNIĆ RAM
    # obiekt stats posiada bardzo pożytyczene metody i przechowuje fitnes
    # osobników z każdej generacji z podziałem na gatunki
    stats = neat.StatisticsReporter()
    p.add_reporter(stats)

    # to zapisuje checkpointy co okreslona liczbe generacji lub co okreslony
    # czas w sekundach lub oba ( co szybciej wyskoczy )
    # p.add_reporter(neat.Checkpointer(5))

    # run to run
    if useParallel:
        pe = neat.ParallelEvaluator(hmWorkers, eval_parallel)
        winner = p.run(pe.evaluate, n=hmGen)
    else:
        winner = p.run(eval_notParallel, n=hmGen)

    # Display the winning genome.
    print('\nBest genome:\n{!s}'.format(winner))

    # Show output of the most fit genome against training data.
    print('\nOutput:')
    winner_net = neat.FeedForwardNetwork.create(winner, config)

    #visualize.draw_net(config, winner, True)
    #visualize.plot_stats(stats, ylog=False, view=True)
    #visualize.plot_species(stats, view=True)

    return winner_net, winner, stats, config

    # przywracanie populacji z checkpointu
    # p = neat.Checkpointer.restore_checkpoint('neat-checkpoint-4')
    # p.run(eval_genomes, 10)


if __name__ == '__main__':
    # Determine path to configuration file. This path manipulation is
    # here so that the script will run successfully regardless of the
    # current working directory.
    local_dir = os.path.dirname(__file__)
    config_path = os.path.join(local_dir, 'config-feedforward')
    winner_net = []
    winner_genome = []
    stats = []
    timeSpans = []
    for k in range(0, hmRun):
        print(k)
        startTime = time.time()
        winner_net_k, winner_genome_k, stats_k, config = run(config_path)
        winner_net.append(winner_net_k)
        winner_genome.append(winner_genome_k)
        stats.append(stats_k)
        endTime = time.time()
        timeSpans.append(endTime - startTime)

# figID = itertools.count(0)
# ploty.plot_average_of_average(stats, hmRun, hmGen, next(figID))
# ploty.plot_average_of_bests(stats, hmRun, hmGen, next(figID))

toSave = [winner_net, winner_genome, stats, target, u, config, timeSpans]
pkl.save_object(toSave, f'zapis_hmGen{hmGen}_hmRun{hmRun}_NEAT')


# bestOutput = netOutput(winner_net, u)
# fig = plt.figure()
# fig.set_size_inches(18.5, 10.5)
# ax = plt.subplot(1, 1, 1)
# ax.grid()
# plt.plot(t, target)
# plt.plot(t, bestOutput)
# plt.show()


