import pickleData as pikl
import numpy as np
import scipy.io


def netOutput(net, u):  # na razie tylko dla SISO bo tak szybciej
    output = [0.0]*len(u)
    for idx, u_k in enumerate(u):  # w kolejnych chwilach, chwil tyle co wymuszenia
        output[idx] = net.activate(u_k)[0]
    return output


def eval_nets(nets, u, target):
    fits = []
    MEs = []
    for net in nets:
        output = np.array(netOutput(net, u)).reshape((-1, 1))
        ME = np.sum(np.abs(output - target)) / hmSample
        NinNet = len(net.node_evals) - 1  # -1 bo nie biorę wyjsciowego
        # DinNet = net.du + net.dy
        fit = 10 - ME - 0.01*NinNet  # - 0.0001*DinNet
        fits.append(fit)
        MEs.append(ME)
    return fits, MEs


mat = scipy.io.loadmat('daneFinal.mat')
Pav = mat["Pav"]
u = mat["x"]
# skalowanie danych
nominalP = 3436*10**6
hmSample = 1100
Pav = Pav / nominalP
target = Pav

data = pikl.load_object("dataCzasDNAS3.pickle")
hmIter = len(data)

bestNets = []
bestGens = []
popStats = []
ME = []
FIT = []
for i in range(0, hmIter):
    bestNets.append(data[i][0])
    bestGens.append(data[i][1])
    popStats.append(data[i][2])
    # srednia me i fit populacji w kolejnych generacjach w każdej generacji osobno
    ME.append(popStats[i].get_ME_mean())
    FIT.append(popStats[i].get_fitness_mean())

# wiersze to kolejne wywołania
# kolumny to kolejne generacje
# to są srednie populacji z
me = np.array(ME)
fit = np.array(FIT)
# srednie me i fit po generacjach we wszystkich iteracjach
meanMeForGen = np.mean(me, axis=0).reshape(1, -1)
meanFitForGen = np.mean(fit, axis=0).reshape(1, -1)

# fity i me najlepszych w ogole uzyskanych osobnikow w każdym wywolaniu algorytmu (iteracji)
bestFit = [x.fitness for x in bestGens]
bestME = [x.ME for x in bestGens]
bestNetIdx = bestFit.index(max(bestFit))
worstNetIdx = bestFit.index(min(bestFit))

bestNet = bestNets[bestNetIdx]
worstNet = bestNets[worstNetIdx]

bestNetOutputUczace = netOutput(bestNet, u)
worstNetOutputUczace = netOutput(worstNet, u)

####################################################
# testowy 1
# mat = scipy.io.loadmat('testowy1.mat')
# hmSample = 1100
# testowy 2
mat = scipy.io.loadmat('testowy2.mat')
hmSample = 1500

Pav = mat["Pav"]
u = mat["x"]
# skalowanie danych
nominalP = 3436*10**6
Pav = Pav / nominalP
target = Pav

bestNetOutputTestowy1 = netOutput(bestNet, u)
worstNetOutputTestowy1 = netOutput(worstNet, u)
# srednia z błedow najlepszych
bestsFitsTest, bestsMEsTest = eval_nets(bestNets, u, target)
print("Średnia z fit osonbików najlepszych")
avgBestFitTest = np.mean(bestsFitsTest)
print(round(avgBestFitTest, 4))

print("Średnia z ME osonbików najlepszych")
avgBestMETest = np.mean(bestsMEsTest)
print(round(avgBestMETest, 4))

# testowy 2
mat = scipy.io.loadmat('testowy2.mat')
u = mat["x"]
bestNetOutputTestowy2 = netOutput(bestNet, u)
worstNetOutputTestowy2 = netOutput(worstNet, u)

# struktura najlepszych
struct = [len(x.nodes)-1 for x in bestGens]


# szukaj max liczby generacji we wszystkich iteracjach
hmGenInIters = [len(x.most_fit_genomes) for x in popStats]
maxGen = max(hmGenInIters)

fitsBestInGen = np.zeros((maxGen, hmIter))
meBestInGen = np.zeros((maxGen, hmIter))
# srednie fity i me najlepszych w kazdej generacji
for generacja in range(0, maxGen):  # po generacjach
    for iteracja in range(0, hmIter):
        # jezeli w tej iteracji nie ma juz takiej generacji
        if generacja >= hmGenInIters[iteracja]:
            # wiersze to generacje
            # kolumny to wywołania programu
            fitsBestInGen[generacja, iteracja] = 20
            meBestInGen[generacja, iteracja] = -1
        else:
            fitsBestInGen[generacja, iteracja] = \
                popStats[iteracja].most_fit_genomes[generacja].fitness
            meBestInGen[generacja, iteracja] = \
                popStats[iteracja].most_fit_genomes[generacja].ME

    # po iteracjach
# fitsBestInGen.append([x.most_fit_genomes[generacja].fitness for x in popStats])
# meBestInGen.append([x.most_fit_genomes[generacja].ME for x in popStats])
# fitsBestInGen = np.array(fitsBestInGen)
# meBestInGen = np.array(meBestInGen)

# liczb srednia po kolumnach
# meanFitsBestInGen = np.mean(fitsBestInGen, 0)
# meanMeBestInGen = np.mean(meBestInGen, 0)

czasy = [x[3] for x in data]
sredniCzas = np.mean(czasy)
dicToSave = {"NEATmeanMeForGen": meanMeForGen,
             "NEATmeanFitForGen": meanFitForGen,
             "NEATbestNetOutputUczace": bestNetOutputUczace,
             "NEATworstNetOutputUczace": worstNetOutputUczace,
             "NEATbestNetOutputTestowy1": bestNetOutputTestowy1,
             "NEATworstNetOutputTestowy1": worstNetOutputTestowy1,
             "NEATbestNetOutputTestowy2": bestNetOutputTestowy2,
             "NEATworstNetOutputTestowy2": worstNetOutputTestowy2,
             "hmHiddenNeurons": struct,
             "bestsFits": bestFit,
             "bestsMEs": bestME,
             "fitsBestInGen": fitsBestInGen,
             "meBestInGen": meBestInGen}

scipy.io.savemat("neatResultsCzasDNAS3.mat", dicToSave)
