import pickleData as pkl
import numpy as np
import matplotlib.pyplot as plt


# pokaz srednia z srednich populacji z wszystkich run
def plot_average_of_average(stats, hmRun, hmGen, ax, lineStyle=None, label=None,
                            flag=True, color=None):
    avgFitRun = np.zeros((hmRun, hmGen))

    for i, s in enumerate(stats):
        run_gens_mean = np.array(s.get_fitness_mean())
        avgFitRun[i, :] = run_gens_mean
    avgFitRun = np.mean(avgFitRun, axis=0)
    #ax = fig.gca()
    if color is None:
        plt.plot(range(0, hmGen, 1), avgFitRun, lineStyle, label=label)
    else:
        plt.plot(range(0, hmGen, 1), avgFitRun, lineStyle, label=label, color=color)
    if flag:
        ax.set_ylim([-500, 0])
        ax.legend()
        ax.xaxis.get_major_locator().set_params(integer=True)
        plt.xlabel(r"Generations $[\#]$")
        plt.ylabel(
            r"Average population fitness from all algorithm calls $[-]$")


def add_subplot_axes(fig, ax, rect):
    box = ax.get_position()
    width = box.width
    height = box.height
    inax_position = ax.transAxes.transform(rect[0:2])
    transFigure = fig.transFigure.inverted()
    infig_position = transFigure.transform(inax_position)
    x = infig_position[0]
    y = infig_position[1]
    width *= rect[2]
    height *= rect[3]  # <= Typo was here
    subax = fig.add_axes([x, y, width, height])  # matplotlib 2.0+
    # subax = fig.add_axes([x, y, width, height], axisbg=axisbg)
    x_labelsize = subax.get_xticklabels()[0].get_size()
    y_labelsize = subax.get_yticklabels()[0].get_size()
    x_labelsize *= rect[2]**0.5
    y_labelsize *= rect[3]**0.5
    subax.xaxis.set_tick_params(labelsize=x_labelsize)
    subax.yaxis.set_tick_params(labelsize=y_labelsize)
    return subax

# pokaz sredni fit najlepszych ze wszystkich run


def plot_average_of_bests(stats, hmRun, hmGen, ax1, lineStyle=None, label=None, color='k'):
    bestsFitRun = np.zeros((hmRun, hmGen))
    bestsMSERun = np.zeros((hmRun, hmGen))
    for i, s in enumerate(stats):
        run_bests_fit = np.array([c.fitness for c in s.most_fit_genomes])
        run_bests_MSE = np.array([c.MSE for c in s.most_fit_genomes])
        bestsFitRun[i, :] = run_bests_fit
        bestsMSERun[i, :] = run_bests_MSE
    avgBestsFit = np.mean(bestsFitRun, axis=0)

    ax1.yaxis.tick_left()
    ax1.yaxis.set_label_position("left")
    ax1.xaxis.get_major_locator().set_params(integer=True)
    ax1.set_ylabel(
        r"Average best individual fitness from all algorithm calls $[-]$")
    ax1.set_xlabel(r"Generations $[\#]$")
    ax1.plot(avgBestsFit, lineStyle, label=label, color=color)
    lines, labels = ax1.get_legend_handles_labels()


plt.rcParams['text.usetex'] = True
plt.rcParams['font.size'] = 30
plt.rcParams['axes.xmargin'] = 0

hmRun = 10
hmGen = 2500

fileName_NEAT = f'zapis_hmGen{hmGen}_hmRun{hmRun}_NEAT'
fileName_myNEAT = f'zapis_hmGen{hmGen}_hmRun{hmRun}_myNEAT'

fileName_NEAT_PWR = fileName_NEAT + '_pwr'
fileName_NEAT_aka = fileName_NEAT + '_aka'
fileName_myNEAT_PWR = fileName_myNEAT + '_pwr'
fileName_myNEAT_aka = fileName_myNEAT + '_aka'

winner_net_NEAT_PWR, winner_genome_NEAT_PWR, stats_NEAT_PWR, target_NEAT_PWR,\
    u_NEAT_PWR, config_NEAT_PWR, timeSpans_NEAT_PWR = pkl.load_object(
        fileName_NEAT_PWR)

winner_net_NEAT_aka, winner_genome_NEAT_aka, stats_NEAT_aka, target_NEAT_aka,\
    u_NEAT_aka, config_NEAT_aka, timeSpans_NEAT_aka = pkl.load_object(
        fileName_NEAT_aka)

winner_net_myNEAT_PWR, winner_genome_myNEAT_PWR, stats_myNEAT_PWR, target_myNEAT_PWR,\
    u_myNEAT_PWR, config_myNEAT_PWR, timeSpans_myNEAT_PWR = pkl.load_object(
        fileName_myNEAT_PWR)

winner_net_myNEAT_aka, winner_genome_myNEAT_aka, stats_myNEAT_aka, target_myNEAT_aka,\
    u_myNEAT_aka, config_myNEAT_aka, timeSpans_myNEAT_aka = pkl.load_object(
        fileName_myNEAT_aka)

# plotuj srednie z populacji
figSrednie = plt.figure(1)
ax = plt.gca()
colors = ['b', 'orange', 'g', 'r']
plot_average_of_average(stats_NEAT_aka, hmRun, hmGen, ax,
                        '-', r'NEAT - application 1', color=colors[0])
plot_average_of_average(stats_myNEAT_aka, hmRun, hmGen, ax,
                        '-', r'dNEAT - application 1', color=colors[1])
plot_average_of_average(stats_NEAT_PWR, hmRun, hmGen, ax,
                        '-', r'NEAT - application 2', color=colors[2])
plot_average_of_average(stats_myNEAT_PWR, hmRun, hmGen, ax,
                        '-', r'dNEAT - application 2', color=colors[3])

rect = [0.07, 0.68, 0.54, 0.3]
axSubplot = add_subplot_axes(figSrednie, ax, rect)
figSrednie = plot_average_of_average(stats_NEAT_aka, hmRun, hmGen, axSubplot,
                                     '-', r'NEAT - application 1', False, color=colors[0])
figSrednie = plot_average_of_average(stats_myNEAT_aka, hmRun, hmGen, axSubplot,
                                     '-', r'dNEAT - application 1', False, color=colors[1])

figSrednie = plot_average_of_average(stats_NEAT_PWR, hmRun, hmGen, axSubplot,
                                     '-', r'NEAT - application 2', False, color=colors[2])
figSrednie = plot_average_of_average(stats_myNEAT_PWR, hmRun, hmGen, axSubplot,
                                     '-', r'dNEAT - application 2', False, color=colors[3])
plt.show()

# plotuj srednie z najlepszych
figBests = plt.figure(2)
ax = plt.gca()
plot_average_of_bests(stats_NEAT_aka, hmRun, hmGen, ax, '--', r'NEAT - application 1', 'r')
plot_average_of_bests(stats_myNEAT_aka, hmRun, hmGen, ax, '--',
                      r'dNEAT - application 1', 'tab:blue')
plot_average_of_bests(stats_NEAT_PWR, hmRun, hmGen, ax, '-', r'NEAT - application 2', 'r')
plot_average_of_bests(stats_myNEAT_PWR, hmRun, hmGen, ax, '-', r'dNEAT - application 2', 'tab:blue')
ax.legend(loc='lower right')
