import numpy as np
import matplotlib.pyplot as plt
import pickleData as pkl
import scipy.io
import visualize_NEAT
import visualize_myNEAT
import myNEAT as mN
import neatSource as Ns
import scipy.io


# pokaz srednia z srednich populacji z wszystkich run
def plot_average_of_average(stats, hmRun, hmGen, fig, lineStyle=None, label=None):
    avgFitRun = np.zeros((hmRun, hmGen))

    for i, s in enumerate(stats):
        run_gens_mean = np.array(s.get_fitness_mean())
        avgFitRun[i, :] = run_gens_mean
    avgFitRun = np.mean(avgFitRun, axis=0)
    ax = fig.gca()
    plt.plot(range(0, hmGen, 1), avgFitRun, lineStyle, label=label)
    ax.legend()
    ax.xaxis.get_major_locator().set_params(integer=True)
    plt.xlabel(r"Generations $[\#]$")
    plt.ylabel(r"Average population fitness from all algorithm calls $[-]$")
    return fig


# pokaz sredni fit najlepszych ze wszystkich run
def plot_average_of_bests(stats, hmRun, hmGen, fig, lineStyle=None, label=None, colors='k'):
    bestsFitRun = np.zeros((hmRun, hmGen))
    bestsMSERun = np.zeros((hmRun, hmGen))
    for i, s in enumerate(stats):
        run_bests_fit = np.array([c.fitness for c in s.most_fit_genomes])
        run_bests_MSE = np.array([c.MSE for c in s.most_fit_genomes])
        bestsFitRun[i, :] = run_bests_fit
        bestsMSERun[i, :] = run_bests_MSE
    avgBestsFit = np.mean(bestsFitRun, axis=0)
    avgBestsMSE = np.mean(bestsMSERun, axis=0)

    if fig.axes == []:
        ax1 = fig.gca()
        ax2 = ax1.twinx()
    else:
        ax1 = fig.axes[0]
        ax2 = fig.axes[1]
    color = colors[0]
    ax1.yaxis.tick_left()
    ax1.yaxis.set_label_position("left")
    ax1.xaxis.get_major_locator().set_params(integer=True)
    ax1.set_ylabel(r"Average best individual fitness from all algorithm calls $[-]$", color=color)
    ax1.set_xlabel(r"Generations $[\#]$")
    ax1.plot(avgBestsFit, lineStyle, label=label+r' - fitness', color=color)
    lines, labels = ax1.get_legend_handles_labels()

    # color = colors[1]
    # ax2.yaxis.tick_right()
    # ax2.yaxis.set_label_position("right")
    # ax2.set_ylabel(r'Average best individual MSE from all algorithm calls $[-]$', color=color)
    # ax2.plot(avgBestsMSE, color=color, linestyle=lineStyle, label=label+r' - MSE')
    # ax2.tick_params(axis='y', labelcolor=color)
    # lines2, labels2 = ax2.get_legend_handles_labels()
    # lines = lines + lines2
    # labels = labels + labels2

    ax2.legend(lines, labels, loc='center right')
    fig.tight_layout()

    return fig


def myNEAT_netOutput(net, u, u0=0, y0=0):  # na razie tylko dla SISO bo tak szybciej
    output = [0.0]*len(u)
    du = [u0] * net.du
    dy = [y0] * net.dy
    for idx, u_k in enumerate(u):  # w kolejnych chwilach, chwil tyle co wymuszenia
        output[idx] = net.activate(list(u_k) + du + dy)[0]
        # jezeli du nie jest puste (self.du != 0)
        if du:
            du.pop(-1)  # usun ostatnie
            du = [u_k[0]] + du  # dodaj u_k
        if dy:
            dy.pop(-1)
            dy = [output[idx]] + dy
    return output


def NEAT_netOutput(net, u):  # na razie tylko dla SISO bo tak szybciej
    output = [0.0]*len(u)
    for idx, u_k in enumerate(u):  # w kolejnych chwilach, chwil tyle co wymuszenia
        output[idx] = net.activate(u_k)[0]
    return output


def plot_best_worst(fig, odpBest, odpWorst, lineStyle, xLabel, label, colors):
    ax1 = fig.gca()
    ax1.yaxis.tick_left()
    ax1.yaxis.set_label_position("left")
    ax1.set_ylabel(r"Responses of ANNs $[-]$")
    ax1.set_xlabel(xLabel)
    plt.plot(t, odpBest, lineStyle, label=label+r' - best', color=colors[0])
    plt.plot(t, odpWorst, lineStyle, label=label+r' - worst', color=colors[1])
    ax1.legend()


# plotuj odp najlepszego i najgorszego po fit
def plot_best_worst_answer(u, t, target, genomes, nets, hmRun, hmGen, fig, lineStyle,
                           label, colors, config, xLabel):
    # znajdz najlpeszeog i najgorszego
    fits = [g.fitness for g in genomes]
    idBest = fits.index(max(fits))
    idWorst = fits.index(min(fits))

    bestGenome = genomes[idBest]
    worstGenome = genomes[idWorst]

    if hasattr(bestGenome, 'du'):
        bestNet = mN.FeedForwardNetwork.create(bestGenome, config)
        worstNet = mN.FeedForwardNetwork.create(worstGenome, config)

        odpBest = np.array(myNEAT_netOutput(bestNet, u)).reshape((-1, 1))
        odpWorst = np.array(myNEAT_netOutput(worstNet, u)).reshape((-1, 1))
    else:
        bestNet = Ns.FeedForwardNetwork.create(bestGenome, config)
        worstNet = Ns.FeedForwardNetwork.create(worstGenome, config)

        odpBest = np.array(NEAT_netOutput(bestNet, u)).reshape((-1, 1))
        odpWorst = np.array(NEAT_netOutput(worstNet, u)).reshape((-1, 1))

    plot_best_worst(fig, odpBest, odpWorst, lineStyle,  xLabel, label, colors)
    return fig, bestGenome, worstGenome, idBest, idWorst


def plotuj2na1(fig, y, y_labels, xAxisLabel, colors, labels=None):
    if labels is None:
        labels = [None]*2
    ax1 = fig.gca()
    ax2 = ax1.twinx()

    ax1.yaxis.tick_left()
    ax1.yaxis.set_label_position("left")
    ax1.xaxis.get_major_locator().set_params(integer=True)
    ax1.set_ylabel(y_labels[0], color=colors[0])
    ax1.set_xlabel(xAxisLabel)
    ax1.plot(y[0], color=colors[0], label=labels[0])
    lines1, labels1 = ax1.get_legend_handles_labels()

    ax2.yaxis.tick_right()
    ax2.yaxis.set_label_position("right")
    ax2.set_ylabel(y_labels[1], color=colors[1])
    ax2.plot(y[1], color=colors[1], label=labels[1])
    ax2.tick_params(axis='y', labelcolor=colors[1])

    lines2, labels2 = ax2.get_legend_handles_labels()
    lines = lines1 + lines2
    labels = labels1 + labels2
    ax2.legend(lines, labels, loc='center right')
    fig.tight_layout()
    plt.show()


plt.rcParams['text.usetex'] = True
plt.rcParams['font.size'] = 30
plt.rcParams['axes.xmargin'] = 0

PWR = False
hmRun = 10
hmGen = 2500

fileName_NEAT = f'zapis_hmGen{hmGen}_hmRun{hmRun}_NEAT'
fileName_myNEAT = f'zapis_hmGen{hmGen}_hmRun{hmRun}_myNEAT'
figSrednie = plt.figure(1)
if PWR:
    xLabel = r'Time $[s]$'
    fileName_NEAT = fileName_NEAT + '_pwr'
    fileName_myNEAT = fileName_myNEAT + '_pwr'
else:
    xLabel = r'Time $[\#]$'
    fileName_NEAT = fileName_NEAT + '_aka'
    fileName_myNEAT = fileName_myNEAT + '_aka'

winner_net_NEAT, winner_genome_NEAT, stats_NEAT, target_NEAT, u_NEAT, config_NEAT, timeSpans_NEAT =\
    pkl.load_object(fileName_NEAT)

winner_net_myNEAT, winner_genome_myNEAT, stats_myNEAT, target_myNEAT, u_myNEAT, config_myNEAT, timeSpans_myNEAT =\
    pkl.load_object(fileName_myNEAT)
t = np.arange(0, len(target_myNEAT), 1)

colors = ['tab:green', 'tab:red']

figSrednie = plt.figure(1)
figSrednie = plot_average_of_average(stats_NEAT, hmRun, hmGen, figSrednie, '--', r'NEAT')
figSrednie = plot_average_of_average(
    stats_myNEAT, hmRun, hmGen, figSrednie, '-', r'dNEAT')

figBests = plt.figure(2)
figBests = plot_average_of_bests(stats_NEAT, hmRun, hmGen, figBests, '--', r'NEAT')
figBests = plot_average_of_bests(stats_myNEAT, hmRun, hmGen, figBests, '-', r'dNEAT')

figBestWorstOdp = plt.figure(3)
plt.plot(t, target_myNEAT, '-', color='tab:blue',
         label='Reference trajectory', linewidth=5, alpha=0.5)

figBestWorstOdp, bestGenome_NEAT, worstGenome_NEAT, idBest_NEAT, idWorst_NEAT =\
    plot_best_worst_answer(u_NEAT, t, target_NEAT, winner_genome_NEAT, winner_net_NEAT,
                           hmRun, hmGen, figBestWorstOdp, '--', r'NEAT',
                           colors, config_NEAT, xLabel)

figBestWorstOdp, bestGenome_myNEAT, worstGenome_myNEAT, idBest_myNEAT, idWorst_myNEAT =\
    plot_best_worst_answer(u_myNEAT, t, target_myNEAT, winner_genome_myNEAT, winner_net_myNEAT,
                           hmRun, hmGen, figBestWorstOdp, '-', r'dNEAT',
                           colors, config_myNEAT, xLabel)
plt.show()
# %%
# plot target i wejscia  uczace
fig = plt.figure(4)
yPlot = [u_myNEAT, target_myNEAT]
colors = ['tab:red', 'tab:blue']
labels = [r"Model input [-]", r"Model output [-]"]
plotuj2na1(fig, yPlot, labels, xLabel, colors)
# pokaz najlepszych
# for i, winner in enumerate(winner_genome_NEAT):
#     visualize_NEAT.draw_net(config_NEAT, winner, True, filename=f'best_NEAT{i}')
# for i, winner in enumerate(winner_genome_myNEAT):
#     visualize_myNEAT.draw_net(config_myNEAT, winner, True, filename=f'best_myNEAT{i}')

# %%
# plotuj odp weryfikacyjne
wer = 2
if PWR:
    # dane
    if wer == 1:
        mat = scipy.io.loadmat('testowy1.mat')
    elif wer == 2:
        mat = scipy.io.loadmat('testowy2.mat')
    Pav = mat["Pav"]
    u = mat["x"]
    # skalowanie danych
    nominalP = 3436*10**6
    hmSample = len(u)
    Pav = Pav / nominalP
    target = Pav

    y0 = target[0]  # nominalna moc
    u0 = -1.098  # nominalne zagłębienie
    t = np.arange(0, hmSample, 1)
else:
    def f(u, y):
        return -0.05*y[1] + 0.02*y[5] + np.sin(y[10]/10) + u[15]
    hmSample = 1085
    t = np.linspace(0, hmSample, hmSample)
    if wer == 1:
        u = np.zeros(t.shape)
        u[25:100] = np.linspace(0, 0.5, 75)
        u[100:150] = 0.5 - 0.1*np.sin(np.linspace(0, 3.14, 50))
        u[150:200] = np.linspace(0.5, 0.75, 50)
        u[200:400] = 0.75-np.linspace(0, 0.5, 200) + 0.25*np.sin(np.linspace(0, 3*3.14, 200))
        u[400:500] = 0
        u[500:600] = 0.25*np.sin(np.linspace(0, 2*3.14, 100))
        u[500:700] = np.linspace(0, 0.8, 200)
        u[700:750] = 0.8
        u[750:900] = np.linspace(0.8, 0, 150) + 0.25*np.sin(np.linspace(0, 2*3.14, 150))
    elif wer == 2:
        # u = u_myNEAT[::-1]
        # u = u.reshape(-1)
        u = np.sin(t/30)
        u[200:400] = 1
        u[400: 600] = 0
        u[600:800] = -1
        u[800:1100] = 1
    y = np.zeros(t.shape)
    maxD = 15+1
    for k in range(maxD, hmSample):
        y[k] = y[k-1] + f(u[k:k-maxD:-1], y[k:k-maxD:-1])

    u = np.array(u[maxD-1:]).reshape((-1, 1))
    target = np.array(y[maxD-1:]).reshape((-1, 1))
    target = target/30.0  # (target - 0)*2 / (32.21283436-0) - 1
    t = np.array(t[maxD-1:])
    hmSample -= maxD
    y0 = 0
    u0 = 0

fig = plt.figure(5)
yPlot = [u, target]
colors = ['tab:red', 'tab:blue']
labels = [r"Model input [-]", r"Model output [-]"]
plotuj2na1(fig, yPlot, labels, xLabel, colors)

bestNet_NEAT = Ns.FeedForwardNetwork.create(bestGenome_NEAT, config_NEAT)
bestOdp_NEAT = np.array(NEAT_netOutput(bestNet_NEAT, u)).reshape((-1, 1))

worstNet_NEAT = Ns.FeedForwardNetwork.create(worstGenome_NEAT, config_NEAT)
worstOdp_NEAT = np.array(NEAT_netOutput(worstNet_NEAT, u)).reshape((-1, 1))

bestNet_myNEAT = mN.FeedForwardNetwork.create(bestGenome_myNEAT, config_myNEAT)
bestOdp_myNEAT = np.array(myNEAT_netOutput(bestNet_myNEAT, u, u0, y0)).reshape((-1, 1))

worstNet_myNEAT = mN.FeedForwardNetwork.create(worstGenome_myNEAT, config_myNEAT)
worstOdp_myNEAT = np.array(myNEAT_netOutput(worstNet_myNEAT, u, u0, y0)).reshape((-1, 1))

fig = plt.figure(6)
plt.plot(target, label=r'Reference trajectory', color='tab:blue', linewidth=5, alpha=0.5)
colors = ['tab:green', 'tab:red']
plot_best_worst(fig, bestOdp_NEAT, worstOdp_NEAT, '--', xLabel, "NEAT", colors)
plot_best_worst(fig, bestOdp_myNEAT, worstOdp_myNEAT, '-', xLabel, "dNEAT", colors)
plt.show()

MSE_best_NEAT_ver = np.sum((bestOdp_NEAT - target)**2) / len(target)
MSE_worst_NEAT_ver = np.sum((worstOdp_NEAT - target)**2) / len(target)

MSE_best_myNEAT_ver = np.sum((bestOdp_myNEAT - target)**2) / len(target)
MSE_worst_myNEAT_ver = np.sum((worstOdp_myNEAT - target)**2) / len(target)

# licz MSE
MSE_traning_NEAT = [g.MSE for g in winner_genome_NEAT]
avg_MSE_traning_NEAT = np.mean(MSE_traning_NEAT)

MSE_traning_myNEAT = [g.MSE for g in winner_genome_myNEAT]
avg_MSE_traning_myNEAT = np.mean(MSE_traning_myNEAT)

MSE_best_NEAT_traning = MSE_traning_NEAT[idBest_NEAT]
MSE_worst_NEAT_traning = MSE_traning_NEAT[idWorst_NEAT]

MSE_best_myNEAT_traning = MSE_traning_myNEAT[idBest_myNEAT]
MSE_worst_myNEAT_traning = MSE_traning_myNEAT[idWorst_myNEAT]

MSE_ver_NEAT = []
for g in winner_genome_NEAT:
    net = Ns.FeedForwardNetwork.create(g, config_NEAT)
    odp = np.array(NEAT_netOutput(net, u)).reshape((-1, 1))
    mse = np.sum((odp - target)**2) / len(target)
    MSE_ver_NEAT.append(mse)
avg_MSE_ver_NEAT = np.mean(MSE_ver_NEAT)

MSE_ver_myNEAT = []
for g in winner_genome_myNEAT:
    net = mN.FeedForwardNetwork.create(g, config_NEAT)
    odp = np.array(myNEAT_netOutput(net, u)).reshape((-1, 1))
    mse = np.sum((odp - target)**2) / len(target)
    MSE_ver_myNEAT.append(mse)
avg_MSE_ver_myNEAT = np.mean(MSE_ver_myNEAT)
